import websockets
import aiomysql
import json
import sys
from datetime import datetime
from typing import List, Mapping
from pymysql import OperationalError
from websockets.exceptions import WebSocketException
from quotesaggregator import settings
from quotesaggregator.utils import logger, Candlestick


class Aggregator:
    """Class responsible for aggregating the data received from the api, generating the candles
    and saving them in the database

    Attributes:

    - :class:`Mapping[int, List[Candlestick]]` __currencies:
    It is a key-value collection where the key is the currency id and the value is a list that has 3 candlesticks,
    the first of which records values every 1 minute, the second every 5 minutes and the last 10 in 10 minutes.
    After the end of each period, the candlestick is saved in the database and their values overwritten.
    """
    __currencies: Mapping[int, List[Candlestick]]

    def __init__(self, loop):
        self.__currencies = {}
        self.loop = loop

    async def save_in_db(self, candle: Candlestick, currency_id: int) -> None:
        """ Saves candlestick in the database

        :param candle: A candlestick with all attributes filled in
        :param currency_id: currency id according to the poloniex table
        """
        try:
            await logger.info("Saving candle")
            conn = await aiomysql.connect(host=settings.DB_HOST, port=int(settings.DB_PORT),
                                          user=settings.DB_USER, password=settings.DB_PASSWORD,
                                          db=settings.DB_NAME, loop=self.loop, autocommit=True)
            async with conn.cursor() as cur:
                # Make SQL query:
                statement = """INSERT INTO `candlestick` VALUES(default, {}, {}, '{}', 
                            {:10.8f}, {:10.8f}, {:10.8f}, {:10.8f})
                            """.format(currency_id, candle.frequency,
                                       candle.date_time.strftime('%Y-%m-%d %H:%M:%S'), candle.open,
                                       candle.low, candle.high, candle.close)
                await cur.execute(statement)
            conn.close()
            await logger.info("Candle saved")
        except TypeError:
            await logger.fatal("The database access configuration has not been defined. Add the environment variables.")
            sys.exit(1)
        except OperationalError as err:
            await logger.fatal(f"An error occurred while accessing the database\n{err.args[0]} - {err.args[1]}")
            exit(1)

    async def run(self) -> None:
        """ Run aggregator """
        logger.info("Program initialized")
        uri = settings.URI
        try:
            async with websockets.connect(uri) as websocket:
                command = settings.COMMAND
                # Send command to websocket to subscribe on channel:
                await websocket.send(command)
                while True:
                    data_str = await websocket.recv()
                    # Parse data:
                    ticker = json.loads(data_str)
                    # Verify if message is an updating a currency:
                    if len(ticker) == 3:
                        # Checks whether the update received is for a chosen currency:
                        if settings.currency_pair.get(ticker[2][0]) is not None:
                            await logger.info("Received value")
                            await self.process_ticker(ticker[2], datetime.now(settings.TIMEZONE))
        except WebSocketException:
            await logger.fatal("Unable to connect to the server, aborted")
            sys.exit(1)

    async def process_ticker(self, ticker: List[int], date_time: datetime) -> None:
        """ It takes the currency value instantly and processes it to be used in a candlestick.

        The value will be saved in the `currencies` attribute if it is a maximum, minimum, opening or closing
        value for a candle.

        :param ticker: An array received from Poloniex API
        :param date_time: Date and time the currency value was received
        """
        # To facilitate the reading and understanding of the code:
        currency_id = ticker[0]
        instant_value = float(ticker[1])
        await logger.info(f"Processing candle, candle {self.__currencies.get(currency_id, None)}")
        # Initially, there will be no candles for the currencies variable, so it will be created:
        if self.__currencies.get(currency_id) is None:
            self.__currencies[currency_id] = [
                # Values are added at the beginning to ensure that all fields are completed and comparable
                # (eg, high, low):
                Candlestick(frequency=1, date_time=date_time, open=instant_value, close=instant_value,
                            low=instant_value, high=instant_value),
                Candlestick(frequency=5, date_time=date_time, open=instant_value, close=instant_value,
                            low=instant_value, high=instant_value),
                Candlestick(frequency=10, date_time=date_time, open=instant_value, close=instant_value,
                            low=instant_value, high=instant_value),
            ]
        candles_frequencies_list = self.__currencies.get(currency_id)
        # Takes the candles of 1, 5, and 10 minutes respectively of the currency received
        for candle in candles_frequencies_list:
            # The candle will only be saved when the cycle ends. Then the rest of the module of the ticker with
            # frequency received must be equal to zero, in addition the minutes cannot be the same.
            # They are not saved before the cycle is closed.
            if date_time.minute != candle.date_time.minute and date_time.minute % candle.frequency == 0:
                minute = candle.date_time.minute - (candle.date_time.minute % candle.frequency)
                candle.date_time = datetime(candle.date_time.year, candle.date_time.month, candle.date_time.day,
                                            candle.date_time.hour, minute)
                frequency = candle.frequency
                open_value = candle.close
                # Creates a copy, this is necessary because copy() method returns a dict, not a Candlestick,
                # in addition, if it is not copied the value is changed before being saved in the database
                await self.save_in_db(Candlestick.from_dict(candle.to_dict()), currency_id)
                # Clean the candle and add close value of the previous candle
                candle.overwrite(frequency, date_time, open_value, open_value, open_value, open_value)

            if candle.high < instant_value:
                candle.high = instant_value
            elif candle.low > instant_value:
                candle.low = instant_value
            # It will be considered the last value of the period
            candle.close = instant_value



import asyncio
from typing import List
from unittest.mock import AsyncMock
import pytest
import mock
from datetime import datetime
from quotesaggregator.utils import Candlestick, logger
from quotesaggregator.aggregator import Aggregator

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


@pytest.fixture(scope='session')
def aggregator_stateful():
    return Aggregator(asyncio.get_event_loop())


@pytest.fixture(scope='function')
def aggregator_stateless():
    return Aggregator(asyncio.get_event_loop())


@mock.patch.object(logger, 'info', new=AsyncMock())
async def test_process_ticker_when_value_is_the_highest(aggregator_stateless: Aggregator):
    """ Checks if it identifies the highest value """
    await aggregator_stateless.process_ticker([1, 1], datetime(2021, 5, 1, 1, 1))
    await aggregator_stateless.process_ticker([1, 3], datetime(2021, 5, 1, 1, 1))
    # As currencies is a private attribute, it must be accessed in this way to be able to test it
    currency_frequencies: List[Candlestick] = aggregator_stateless._Aggregator__currencies.get(1)
    for candle in currency_frequencies:
        assert candle.high == 3.0


@mock.patch.object(logger, 'info', new=AsyncMock())
async def test_process_ticker_when_value_is_the_lowest(aggregator_stateless):
    """ Checks if it identifies the lowest value """
    await aggregator_stateless.process_ticker([1, 2], datetime(2021, 5, 1, 1, 1))
    await aggregator_stateless.process_ticker([1, 3], datetime(2021, 5, 1, 1, 1))
    # As currencies is a private attribute, it must be accessed in this way to be able to test it
    currency_frequencies: List[Candlestick] = aggregator_stateless._Aggregator__currencies.get(1)
    for candle in currency_frequencies:
        assert candle.low == 2.0


@mock.patch.object(logger, 'info', new=AsyncMock())
@mock.patch.object(Aggregator, 'save_in_db', new=AsyncMock())
async def test_process_ticker_when_a_single_value_was_sent_to_the_function(aggregator_stateless: Aggregator):
    """ Checks whether the candle is created and added to currencies """
    await aggregator_stateless.process_ticker([1, 2], datetime(2021, 5, 1, 1, 1))
    # As currencies is a private attribute, it must be accessed in this way to be able to test it
    currency_frequencies = aggregator_stateless._Aggregator__currencies.get(1)
    assert currency_frequencies[0].to_dict() == Candlestick(frequency=1, date_time=datetime(2021, 5, 1, 1, 1),
                                                            open=2, close=2, low=2, high=2).to_dict()
    assert currency_frequencies[1].to_dict() == Candlestick(frequency=5, date_time=datetime(2021, 5, 1, 1, 1),
                                                            open=2, close=2, low=2, high=2).to_dict()
    assert currency_frequencies[2].to_dict() == Candlestick(frequency=10, date_time=datetime(2021, 5, 1, 1, 1),
                                                            open=2, close=2, low=2, high=2).to_dict()


@mock.patch.object(logger, 'info', new=AsyncMock())
@mock.patch.object(Aggregator, 'save_in_db', new=AsyncMock())
@pytest.mark.parametrize("value, date_time", [
    ([0, 2], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([0, 1], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([0, 4], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([0, 3], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([0, 1], datetime(year=2021, month=5, day=1, hour=1, minute=1)),
])
async def test_process_ticker_forming_a_one_minute_candle(value: list, date_time: datetime,
                                                          aggregator_stateful: Aggregator):
    """ Checks whether the one-minute candle is valid and sent to be saved """
    await aggregator_stateful.process_ticker(value, date_time)
    aggregator_stateful.save_in_db: AsyncMock
    if aggregator_stateful.save_in_db.called:
        aggregator_stateful.save_in_db.assert_called_with(
            Candlestick(frequency=1, date_time=datetime(year=2021, month=5, day=1, hour=1, minute=0), open=2, close=3,
                        low=1, high=4), 0)


@mock.patch.object(logger, 'info', new=AsyncMock())
@mock.patch.object(Aggregator, 'save_in_db', new=AsyncMock())
@pytest.mark.parametrize("value, date_time", [
    ([2, 2], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([2, 2], datetime(year=2021, month=5, day=1, hour=1, minute=1)),
    ([2, 1], datetime(year=2021, month=5, day=1, hour=1, minute=2)),
    ([2, 4], datetime(year=2021, month=5, day=1, hour=1, minute=3)),
    ([2, 3], datetime(year=2021, month=5, day=1, hour=1, minute=4)),
    ([2, 1], datetime(year=2021, month=5, day=1, hour=1, minute=5)),
])
async def test_process_ticker_forming_a_five_minute_candle(value: list, date_time: datetime,
                                                           aggregator_stateful: Aggregator):
    """ Checks whether the five minute candle is valid and sent to be saved """
    await aggregator_stateful.process_ticker(value, date_time)
    aggregator_stateful.save_in_db: AsyncMock
    if aggregator_stateful.save_in_db.called:
        # there are calls from other cycles that we are not interested in for the test
        if aggregator_stateful.save_in_db.call_count % 5 == 0:
            aggregator_stateful.save_in_db.assert_called_with(
                Candlestick(frequency=5, date_time=datetime(year=2021, month=5, day=1, hour=1, minute=0), open=2,
                            close=3,
                            low=1, high=4), 1)


@mock.patch.object(logger, 'info', new=AsyncMock())
@mock.patch.object(Aggregator, 'save_in_db', new=AsyncMock())
@pytest.mark.parametrize("value, date_time", [
    ([3, 2], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
    ([3, 2], datetime(year=2021, month=5, day=1, hour=1, minute=1)),
    ([3, 1], datetime(year=2021, month=5, day=1, hour=1, minute=2)),
    ([3, 4], datetime(year=2021, month=5, day=1, hour=1, minute=3)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=4)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=6)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=7)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=8)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=9)),
    ([3, 3], datetime(year=2021, month=5, day=1, hour=1, minute=10)),
])
async def test_process_ticker_forming_a_ten_minute_candle(value: list, date_time: datetime,
                                                          aggregator_stateful: Aggregator):
    """ Checks whether the ten minute candle is valid and sent to be saved """
    await aggregator_stateful.process_ticker(value, date_time)
    aggregator_stateful.save_in_db: AsyncMock
    if aggregator_stateful.save_in_db.called:
        # there are calls from other cycles that we are not interested in for the test
        if aggregator_stateful.save_in_db.call_count % 10 == 0:
            aggregator_stateful.save_in_db.assert_called_with(
                Candlestick(frequency=10, date_time=datetime(year=2021, month=5, day=1, hour=1, minute=0), open=2,
                            close=3,
                            low=1, high=4), 3)


@mock.patch.object(logger, 'info', new=AsyncMock())
@mock.patch.object(Aggregator, 'save_in_db', new=AsyncMock())
@pytest.mark.parametrize("value, date_time", [
    ([4, 2], datetime(year=2021, month=5, day=1, hour=1, minute=51)),
    ([4, 2], datetime(year=2021, month=5, day=1, hour=1, minute=52)),
    ([4, 1], datetime(year=2021, month=5, day=1, hour=1, minute=53)),
    ([4, 4], datetime(year=2021, month=5, day=1, hour=1, minute=54)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=55)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=56)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=57)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=58)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=59)),
    ([4, 3], datetime(year=2021, month=5, day=1, hour=1, minute=0)),
])
async def test_process_ticker_forming_a_multiple_candles(value: list, date_time: datetime,
                                                         aggregator_stateful: Aggregator):
    """ Checks whether the ten minute candle is valid and sent to be saved """
    await aggregator_stateful.process_ticker(value, date_time)
    aggregator_stateful.save_in_db: AsyncMock
    if aggregator_stateful.save_in_db.called:
        if aggregator_stateful.save_in_db.call_args.args[0].frequency == 1:
            pass
        elif aggregator_stateful.save_in_db.call_count % 5 == 0:
            assert aggregator_stateful.save_in_db.call_args.args[0].frequency == 5
        elif aggregator_stateful.save_in_db.call_count % 10 == 0:
            aggregator_stateful.save_in_db.assert_called_with(
                Candlestick(frequency=10, date_time=datetime(year=2021, month=5, day=1, hour=1, minute=0), open=2,
                            close=3,
                            low=1, high=4), 3)


if __name__ == '__main__':
    pytest.main()


def run():
    pytest.main()



from prodict import Prodict
from datetime import datetime
import aiologger


class Candlestick(Prodict):
    """ An object similar to a dictionary to make it easier to read the code """
    frequency: int
    date_time: datetime
    open: float
    close: float
    low: float
    high: float

    def overwrite(self, frequency: int, date_time: datetime, open: float, close: float, low: float, high: float):
        """ Overwrite attributes """
        self.frequency = frequency
        self.date_time = date_time
        self.open = open
        self.close = close
        self.low = low
        self.high = high


# Logger
logger = aiologger.Logger.with_default_handlers(level=10)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['quotesaggregator', 'quotesaggregator.tests']

package_data = \
{'': ['*']}

install_requires = \
['aiologger>=0.6.1,<0.7.0',
 'aiomysql>=0.0.21,<0.0.22',
 'asyncio>=3.4.3,<4.0.0',
 'cryptography>=3.4.7,<4.0.0',
 'prodict>=0.8.16,<0.9.0',
 'pytz>=2021.1,<2022.0',
 'websockets>=9.0.1,<10.0.0']

entry_points = \
{'console_scripts': ['aggregator-run = '
                     'quotesaggregator.__main__:run_forever_async',
                     'aggregator-run-tests = '
                     'quotesaggregator.tests.test_aggregator:run']}

setup_kwargs = {
    'name': 'quotesaggregator',
    'version': '0.1.0',
    'description': 'QuotesAggregator is a quote aggregator program that stores the instant values of various currencies to create and save 1, 5 and 10 minute candles in a database through the Poloniex API.',
    'long_description': '# QuotesAggregator\n\nQuotesAggregator é um programa agregador de cotações que armazena os valores instantâneos de várias moedas para criar e\nsalvar candles de 1, 5 e 10 minutos num banco de dados através da API da Poloniex.\n\n- Gráficos gerados a partir do banco de dados:\n\n![Gráfico de 1 minuto](.github/chart1.png)\n![Gráfico de 5 minutos](.github/chart5.png)\n![Gráfico de 10 minutos](.github/chart10.png)\n\n## Como Utilizar\n\n### Pré-requisitos\n\n- Docker\n\n### Utilizando\n\n```shell\ndocker-compose up\n```\n\n#### Executando testes unitários\n\n```shell\n# Na pasta raiz do projeto\n# Lembre-se de criar um virtual env\npip install --no-cache-dir -r requirements.txt\npytest\n```\n\n### Criando graficos\n\n[Clique aqui](createcharts/README.md)\n\n## Decisões de arquitetura\n\n- Uso de websokets ao invés da HTTP API\n    - Para manter a consistência dos dados seria necessário fazer varias requisições a HTTP API e por conta disso,\n      poderia ser bloqueado ou ter requisições recusadas, além de receber dados desnecessários. Por conta disso,\n      utilizei websockets API, dessa forma basta inscrever-se no canal para receber a atualização das moedas.\n- Uso de bibliotecas asyncio\n    - O programa demanda muito IO (receber os dados da API, salva-los), por isso, optei por utilizar bibliotecas\n      assíncronas para isso, dessa forma o programa não é bloqueado enquanto a recebimento ou envio de dados.\n- Utilização do prodict\n    - Para facilitar a leitura do código, utilizei uma classe que se comporta como um dicionário, porém deixa os\n      atributos mais legíveis e funcionalidade de autocomplete.\n- Objeto candle não contem a currency_id\n    - Uma estratégia possível, seria utilizar outra chave do candle contendo o ‘id’ da moeda, porém ao utiliza-lo seria\n      necessário iterar, no pior caso, sobre toda a lista, ou seja, O(n). Utilizando um dicionário que se comporta como\n      um HashMap a complexidade de tempo cai para [O(1) no caso médio](https://wiki.python.org/moin/TimeComplexity#dict)\n      , ou seja, a complexidade de tempo para encontrar os candles é O(1).\n- Não Utilização de ORM\n    - Também seria possível utilizar um ORM para facilitar a interação com o banco de dados, porém optei por utilizar\n      consultas SQL para demonstrar os meus conhecimentos em SQL (Mesmo que, nessa prova, apenas consultas simples são\n      necessárias).\n\n### Funcionamento\n\nApos se inscrever no canal Ticker Data é recebido a atualização do valor das moedas, este valor é processado e a partir\ndele é criado um objeto chave-valor que contem uma lista de 3 candles de 1 5, 10 minutos respetivamente, então estes\ncandles recebem atualização constantemente, até que o período do candle se encerre e ele seja salvo e os seus atributos\nsobrescritos.\n\n### Resultados\n\nOs candles gerados podem ser encontrados através do banco de dados disponível na porta `34807` da sua maquina. Além\ndisso, o programa gera logs visíveis no stdout do docker.\n\n### Dificuldades (Resolvidas)\n\n- Como nunca havia testado métodos assíncronos ainda, foi difícil entender como faze-lo.\n- Ao criar a tabela, utilizei float para os campos, e não comportava o tamanho de alguns valores recebidos, mudei para\n  DECIMAL, que inclusive é mais adequado para valores monetários por problemas de arredondamento em outros tipos de\n  dados.\n- Por desconhecer o modulo aiomysql, cometi o erro de não fazer o commit na transação do banco de dados, e por isso, os\n  dados não eram salvos. Para corrigir habilitei o autocommit na chamada.\n- Tentei algumas abordagens para saber quando salvar o candle, uma delas deixava os valores de abertura-fechamento\n  errado, pois ele considerava o valor pertencente ao período como valor inicial, sendo que o valor do final de um deve\n  ser igual ao inicial do outro, além disso, o candle de 1 minuto estava sendo atualizado a cada 2 minutos, pois eu\n  utilizei o modulo de 2 ao invés do de 1, porque todo numero dividido por um tem resto 0, então ele salvaria o candle\n  antes de o minuto ser finalizado. A solução foi junto ao modulo, verificar se o tempo do candle atual era diferente do\n  novo valor recebido.',
    'author': 'Lucas',
    'author_email': 'lmr2199@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mrlucasrib/QuotesAggregator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import asyncio
from quotesaggregator import aggregator


def run_forever_async():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(aggregator.Aggregator(loop).run())


if __name__ == '__main__':
    run_forever_async()
